#!/usr/bin/env python

# =============================================================================
# MODULE DOCSTRING
# =============================================================================

"""
Classes and utilities for multi-task collective variable.
"""

__all__ = ["MultiTaskCV"]

# =============================================================================
# GLOBAL IMPORTS
# =============================================================================

from typing import Optional, Sequence

import torch

from mlcolvar.cvs.cv import BaseCV


# =============================================================================
# VARIATIONAL AUTOENCODER CV
# =============================================================================


class MultiTaskCV:
    """Multi-task collective variable.

    This class wraps an existing CV object and adds a linear combination of other
    auxiliary loss functions that target different datasets with different information.
    The class works only if the ``main_cv`` does not make use of ``__slots__``.

    Examples
    --------

    A semi-supervised autoencoder mixing ELBO and the Fisher's discriminant loss.

    >>> from mlcolvar.cvs import AutoEncoderCV
    >>> from mlcolvar.core.loss import FisherDiscriminantLoss
    >>> from mlcolvar.data import DictDataset, DictModule

    >>> n_descriptors = 5
    >>> n_labels = 2  # Number of states
    >>> n_cvs = 2

    Initialize the multi-task CV. The Fisher's discriminant loss has half the
    weight of the ELBO loss.

    >>> main_cv = AutoEncoderCV(encoder_layers=[n_descriptors, 10, n_cvs])
    >>> aux_loss_fn = FisherDiscriminantLoss(n_states=n_labels)
    >>> multi_cv = MultiTaskCV(main_cv, auxiliary_loss_fns=[aux_loss_fn], loss_coefficients=[0.5])

    MultiTaskCV now exposes the same API as AutoEncoderCV.

    >>> multi_cv.norm_in.set_custom(mean=torch.tensor(0.0), range=torch.tensor(1.0))

    Create a multi-dataset datamodule for this CV.

    >>> n_samples = 100
    >>> unsupervised_dataset = DictDataset({
    ...     'data': torch.rand(n_samples, n_descriptors),
    ... })
    >>> supervised_dataset = DictDataset({
    ...     'data': torch.rand(n_samples, n_descriptors),
    ...     'labels': torch.tensor([0., 1]).repeat(n_samples//2)
    ... })
    >>> datamodule = DictModule(dataset=[unsupervised_dataset, supervised_dataset])

    # Create a PyTorch Lightning trainer.
    >>> import lightning
    >>> trainer = lightning.Trainer(max_epochs=1, log_every_n_steps=5, logger=None, enable_checkpointing=False)

    """

    def __init__(
        self,
        main_cv: BaseCV,
        auxiliary_loss_fns: Sequence,
        loss_coefficients: Optional[Sequence[float]] = None,
    ):
        """
        Constructor.

        Parameters
        ----------
        main_cv : BaseCV
            The main collective variable. The CV will dynamically inherit from
            this object's class and expose all its members.
        auxiliary_loss_fns : list
            A list of auxiliary loss functions.
        loss_coefficients : list-like of floats, optional
            A list of length ``len(auxiliary_loss_fns)`` with the coefficients
            of the linear combination of loss functions. If not provided, all
            auxiliary loss functions are assigned coefficient 1 (the main CV
            has always coefficient 1).

        """
        # This changes dynamically the class of this object to inherit both from
        # MultiTaskCV and main_cv.__class__ so that we can access all members of
        # main_cv and still be able to override some of them.
        self.__class__ = type(
            "MultiTask" + main_cv.__class__.__name__,  # class name
            (self.__class__, main_cv.__class__),  # base classes
            {},  # self.__class__.__dict__
        )

        # Copy all members of main_cv into this object.
        self.__dict__ = main_cv.__dict__

        self.auxiliary_loss_fns = torch.nn.ModuleList(auxiliary_loss_fns)
        self.loss_coefficients = loss_coefficients

    def training_step(self, train_batch, batch_idx):
        stage = "train" if self.training else "valid"

        # Compute main loss (the main CV should already log the first loss).
        loss = super().training_step(train_batch["dataset0"], batch_idx)

        # Compute auxiliary losses one by one.
        for loss_idx, aux_loss_fn in enumerate(self.auxiliary_loss_fns):
            dataset_batch = train_batch["dataset" + str(loss_idx + 1)]

            # Prepare keyword arguments to pass to the auxiliary loss function.
            aux_loss_kwargs = {
                k: v for k, v in dataset_batch.items() if not k.startswith("data")
            }

            # Forward data of this dataset (and eventually the time-lagged one).
            cv = self.forward_cv(dataset_batch["data"])
            try:
                cv_lag = self.forward_cv(dataset_batch["data_lag"])
            except KeyError:  # Not a time-lagged CV.
                aux_loss = aux_loss_fn(cv, **aux_loss_kwargs)
            else:
                aux_loss = aux_loss_fn(cv, cv_lag, **aux_loss_kwargs)

            # Log the auxiliary loss (before the coefficient).
            self.log(f"{stage}_aux_loss_{loss_idx}", aux_loss.to(float), on_epoch=True)

            if self.loss_coefficients is not None:
                aux_loss = self.loss_coefficients[loss_idx] * aux_loss
            loss = loss + aux_loss

        # Log the total loss
        self.log(f"{stage}_total_loss", loss.to(float), on_epoch=True)

        # return loss
        return loss


if __name__ == "__main__":
    import doctest

    doctest.testmod()



#!/usr/bin/env python

# =============================================================================
# MODULE DOCSTRING
# =============================================================================

"""
Tests for the members of the mlcolvar.data.dataloader module.
"""


# =============================================================================
# GLOBAL IMPORTS
# =============================================================================

import math

import pytest
import torch
from torch.utils.data import Subset

from mlcolvar.data.dataset import DictDataset
from mlcolvar.data.dataloader import DictLoader


# =============================================================================
# UTILITY FUNCTIONS
# =============================================================================


def create_random_datasets(same_batch_size=True):
    """A list of datasets with different keys and different number of samples."""
    # One number of samples for each of the dataset.
    if same_batch_size:
        n_samples = [10, 10, 10]
    else:
        n_samples = [i * 10 for i in (1, 2, 3)]

    # Create datasets.
    datasets = [
        # Subset of a DictDataset.
        Subset(
            DictDataset({"data": torch.randn(n_samples[0] + 2, 2)}),
            indices=list(range(1, n_samples[0] + 1)),
        ),
        # DictDataset.
        DictDataset(
            {"data": torch.randn(n_samples[1], 2), "labels": torch.randn(n_samples[1])}
        ),
        # Standard dictionary.
        {
            "data": torch.randn(n_samples[2], 2),
            "labels": torch.randn(n_samples[2]),
            "weights": torch.randn(n_samples[2]),
        },
    ]
    return datasets


# =============================================================================
# TESTS
# =============================================================================


def test_fast_dictionary_loader_init():
    """DictLoader can be initialized from dict, DictDataset, and Subsets."""
    x = torch.arange(1, 11)
    y = x**2
    x = x.unsqueeze(1)

    batch_size = 2

    def _check_dataloader(dl, n_samples=10, n_batches=5, shuffled=False):
        assert dl.dataset_len == n_samples
        assert len(dl) == n_batches

        # The detaset is converted to a DictDataset.
        assert isinstance(dl.dataset, DictDataset)

        # Check the batch.
        batch = next(iter(dl))
        assert set(batch.keys()) == set(["data", "labels"])
        if not shuffled:
            assert torch.all(batch["data"] == torch.tensor([[1.0], [2.0]]))
            assert torch.all(batch["labels"] == torch.tensor([1.0, 4.0]))

        # Count number of iterated batches.
        counter = 0
        for batch in dl:
            counter += 1
        assert counter == n_batches

    # Start from dictionary
    d = {"data": x, "labels": y}
    dataloader = DictLoader(d, batch_size=batch_size, shuffle=False)
    _check_dataloader(dataloader)

    # or from DictDataset
    dict_dataset = DictDataset(d)
    dataloader = DictLoader(dict_dataset, batch_size=batch_size, shuffle=False)
    _check_dataloader(dataloader)

    # or from subset
    train, _ = torch.utils.data.random_split(dict_dataset, [0.5, 0.5])
    dataloader = DictLoader(train, batch_size=batch_size, shuffle=False)
    _check_dataloader(dataloader, n_samples=5, n_batches=3, shuffled=True)

    # an error is raised if initialized from anything else.
    dataset = torch.utils.data.TensorDataset(x)
    with pytest.raises(ValueError, match="must be of type"):
        DictLoader(dataset)


@pytest.mark.parametrize("same_batch_size", [True, False])
@pytest.mark.parametrize("shuffle", [True, False])
def test_fast_dictionary_loader_multidataset_batch(same_batch_size, shuffle):
    """DictLoader combines multiple datasets into a single batch."""
    datasets = create_random_datasets(same_batch_size)
    n_datasets = len(datasets)

    # Create the dataloader.
    if same_batch_size:
        batch_size = 2
        batch_size_list = [batch_size for _ in range(n_datasets)]
    else:
        # The second/third datasets have twice/thrice many samples.
        batch_size = [2, 4, 6]
        batch_size_list = batch_size
    n_batches = math.ceil(len(datasets[0]) / batch_size_list[0])
    dataloader = DictLoader(datasets, batch_size=batch_size, shuffle=True)

    # Check that dataset_len, number of batches, and batch sizes are computed correctly.
    assert dataloader.dataset_len == [len(d) for d in dataloader.dataset]
    assert len(dataloader) == n_batches
    assert dataloader.batch_size == batch_size_list

    # Test that the batches are correct.
    batch_counter = 0
    for batch in dataloader:
        batch_counter += 1

        # batch has one key for each dataset.
        assert len(batch) == n_datasets

        # Check that the batches have the expected number of keys created in create_random_datasets().
        for i in range(n_datasets):
            n_dataset_keys = len(batch[f"dataset{i}"])
            assert n_dataset_keys == i + 1

        # Check shape 'data' (all datasets from create_random_datasets() have this key).
        for i in range(n_datasets):
            assert batch[f"dataset{i}"]["data"].shape == (batch_size_list[i], 2)

        # Check shape 'labels' (only the last two datasets have it).
        for i in range(1, n_datasets):
            assert batch[f"dataset{i}"]["labels"].shape == (batch_size_list[i],)

        # Check shape 'weights' (only the last dataset has it).
        assert batch[f"dataset{i}"]["weights"].shape == (batch_size_list[i],)

    # Iterator returns correct number of batches.
    assert batch_counter == n_batches


def test_fast_dictionary_loader_multidataset_different_n_batches():
    """DictLoader complains if the number of batches for different datasets is different."""
    # Two datasets with same or different number of samples.
    n_samples = 10
    datasets_same = [
        {"data": torch.randn(n_samples, 2)},
        {"data": torch.randn(n_samples, 2)},
    ]
    datasets_different = [
        {"data": torch.randn(n_samples, 2)},
        {"data": torch.randn(n_samples * 2, 2)},
    ]

    # Test error raising.
    err_msg = "must have the same number of batches"
    with pytest.raises(ValueError, match=err_msg):
        DictLoader(datasets_different, batch_size=1)
    with pytest.raises(ValueError, match=err_msg):
        DictLoader(datasets_different, batch_size=[2, 2])

    # Same number of samples but different batch sizes.
    with pytest.raises(ValueError, match=err_msg):
        DictLoader(datasets_same, batch_size=[1, 2])

    # Same error if we try to set separately incompatible datasets or batch_sizes.
    loader = DictLoader(datasets_same, batch_size=2)
    with pytest.raises(ValueError, match=err_msg):
        loader.dataset = datasets_different
    with pytest.raises(ValueError, match=err_msg):
        loader.batch_size = [2, 4]

    # While setting them together works.
    loader.set_dataset_and_batch_size(dataset=datasets_different, batch_size=[2, 4])
    assert loader.batch_size == [2, 4]


def test_fast_dictionary_loader_multidataset_get_stats():
    """DictLoader compute stats for all or a subset of datasets."""
    datasets = create_random_datasets()
    dataloader = DictLoader(datasets)

    # Stats for all datasets.
    stats_all = dataloader.get_stats()
    assert tuple(stats_all.keys()) == tuple(f"dataset{i}" for i in range(len(datasets)))

    # Check that statistics for all the datasets's key have been computed.
    for dataset_idx, dataset_keys in enumerate(dataloader.keys):
        dataset_name = "dataset" + str(dataset_idx)
        assert tuple(stats_all[dataset_name].keys()) == dataset_keys

    # Stats for single datasets.
    stats_1 = dataloader.get_stats(dataset_idx=1)

    # Check that the statistics for dataset1 agree for both implementations.
    for dataset_key in dataloader.keys[1]:
        for stat_name, stat in stats_1[dataset_key].items():
            assert torch.allclose(stat, stats_all["dataset1"][dataset_key][stat_name])



##########################################################################
## FESSA COLOR PALETTE
##########################################################################
#  https://github.com/luigibonati/fessa-color-palette/blob/master/fessa.py

import mlcolvar
from matplotlib.colors import LinearSegmentedColormap, ColorConverter
from matplotlib import patches as mpatches
import matplotlib as mpl

__all__ = ["paletteFessa", "paletteCortina"]

# Fessa colormap
paletteFessa = [
    "#1F3B73",  # dark-blue
    "#2F9294",  # green-blue
    "#50B28D",  # green
    "#A7D655",  # pisello
    "#FFE03E",  # yellow
    "#FFA955",  # orange
    "#D6573B",  # red
]

cm_fessa = LinearSegmentedColormap.from_list("fessa", paletteFessa)
mpl.colormaps.register(cmap=cm_fessa)
mpl.colormaps.register(cmap=cm_fessa.reversed())

for i in range(len(paletteFessa)):
    ColorConverter.colors[f"fessa{i}"] = paletteFessa[i]

### To set it as default
# import fessa
# plt.set_cmap('fessa')
### or the reversed one
# plt.set_cmap('fessa_r')
### For contour plots
# plt.contourf(X, Y, Z, cmap='fessa')
### For standard plots
# plt.plot(x, y, color='fessa0')


# Cortina1980 colormap
paletteCortina = [
    [0.0, 0.0, 0.803921568627451, 1],  # mediumblue
    [0.4823529411764706, 0.40784313725490196, 0.9333333333333333, 1],  # mediumslateblue
    [0.0, 0.9803921568627451, 0.6039215686274509, 1],  # mediumspringgreen
    [0.23529411764705882, 0.7019607843137254, 0.44313725490196076, 1],  # mediumseagreen
    [0.8588235294117647, 0.4392156862745098, 0.5764705882352941, 1],  # palevioletred
    [0.7803921568627451, 0.08235294117647059, 0.5215686274509804, 1],  # mediumvioletred
]

cm_cortina = LinearSegmentedColormap.from_list("cortina80", paletteCortina)
mpl.colormaps.register(cmap=cm_cortina)
mpl.colormaps.register(cmap=cm_cortina.reversed())

##########################################################################
## HELPER FUNCTIONS FOR 2D SYSTEMS
##########################################################################

import matplotlib.pyplot as plt
import numpy as np
import torch


def muller_brown_potential(x, y):
    """Muller-Brown analytical potential"""
    prefactor = 0.15
    A = (-200, -100, -170, 15)
    a = (-1, -1, -6.5, 0.7)
    b = (0, 0, 11, 0.6)
    c = (-10, -10, -6.5, 0.7)
    x0 = (1, 0, -0.5, -1)
    y0 = (0, 0.5, 1.5, 1)
    offset = -146.7

    v = -prefactor * offset
    for i in range(4):
        v += (
            prefactor
            * A[i]
            * np.exp(
                a[i] * (x - x0[i]) ** 2
                + b[i] * (x - x0[i]) * (y - y0[i])
                + c[i] * (y - y0[i]) ** 2
            )
        )
    return v


def muller_brown_mfep():
    mfep = np.loadtxt(
        mlcolvar.__path__[0]
        + "/../docs/notebooks/tutorials/data/muller-brown/mfep.txt",
        usecols=(0, 1),
    )
    return mfep


def muller_brown_potential_three_states(x, y):
    """Muller-Brown analytical potential"""
    prefactor = 0.15
    A = (-280, -170, -170, 15)
    a = (-15, -1, -6.5, 0.7)
    b = (0, 0, 11, 0.6)
    c = (-10, -10, -6.5, 0.7)
    x0 = (1, 0.2, -0.5, -1)
    y0 = (0, 0.5, 1.5, 1)
    offset = -146.7

    v = -prefactor * offset
    for i in range(4):
        v += (
            prefactor
            * A[i]
            * np.exp(
                a[i] * (x - x0[i]) ** 2
                + b[i] * (x - x0[i]) * (y - y0[i])
                + c[i] * (y - y0[i]) ** 2
            )
        )
    return v


def muller_brown_three_states_mfep():
    mfep = np.loadtxt(
        mlcolvar.__path__[0]
        + "/../docs/notebooks/tutorials/data/muller-brown-3states/mfep.txt",
        usecols=(0, 1),
    )
    return mfep


def plot_isolines_2D(
    function,
    component=None,
    limits=((-1.8, 1.2), (-0.4, 2.1)),
    num_points=(100, 100),
    mode="contourf",
    levels=12,
    cmap=None,
    colorbar=None,
    max_value=None,
    ax=None,
    **kwargs,
):
    """Plot isolines of a function/model in a 2D space."""

    # Define grid where to evaluate function
    if type(num_points) == int:
        num_points = (num_points, num_points)
    xx = np.linspace(limits[0][0], limits[0][1], num_points[0])
    yy = np.linspace(limits[1][0], limits[1][1], num_points[1])
    xv, yv = np.meshgrid(xx, yy)

    # if torch module
    if isinstance(function, torch.nn.Module):
        z = np.zeros_like(xv)
        for i in range(num_points[0]):
            for j in range(num_points[1]):
                xy = torch.Tensor([xv[i, j], yv[i, j]])
                with torch.no_grad():
                    train_mode = function.training
                    function.eval()
                    s = function(xy).numpy()
                    function.training = train_mode
                    if component is not None:
                        s = s[component]
                z[i, j] = s
    # else apply function directly to grid points
    else:
        z = function(xv, yv)

    if max_value is not None:
        z[z > max_value] = max_value

    # Setup plot
    return_axs = False
    if ax is None:
        return_axs = True
        _, ax = plt.subplots(figsize=(6, 4.0), dpi=100)

    # Color scheme
    if cmap is None:
        if mode == "contourf":
            cmap = "fessa"
        elif mode == "contour":
            cmap = "Greys_r"

    # Colorbar
    if colorbar is None:
        if mode == "contourf":
            colorbar = True
        elif mode == "contour":
            colorbar = False

    # Plot
    if mode == "contourf":
        pp = ax.contourf(xv, yv, z, levels=levels, cmap=cmap, **kwargs)
        if colorbar:
            plt.colorbar(pp, ax=ax)
    else:
        pp = ax.contour(xv, yv, z, levels=levels, cmap=cmap, **kwargs)

    if return_axs:
        return ax
    else:
        return None


def plot_metrics(
    metrics,
    keys=["train_loss_epoch", "valid_loss"],
    x=None,  # 'epoch'
    labels=None,  # ['Train','Valid'],
    linestyles=None,  # ['-','--']
    colors=None,  # ['fessa0','fessa1']
    xlabel="Epoch",
    ylabel="Loss",
    title="Learning curves",
    yscale=None,
    ax=None,
):
    """Plot logged metrics."""

    # Setup axis
    return_axs = False
    if ax is None:
        return_axs = True
        _, ax = plt.subplots(figsize=(5, 4), dpi=100)

    # Plot metrics
    auto_x = True if x is None else False
    for i, key in enumerate(keys):
        y = metrics[key]
        lstyle = linestyles[i] if linestyles is not None else None
        label = labels[i] if labels is not None else key
        color = colors[i] if colors is not None else None
        if auto_x:
            x = np.arange(len(y))
        ax.plot(x, y, linestyle=lstyle, label=label, color=color)

    # Plot settings
    if xlabel is not None:
        ax.set_xlabel(xlabel)
    if ylabel is not None:
        ax.set_ylabel(ylabel)
    if title is not None:
        ax.set_title(title)
    if yscale is not None:
        ax.set_yscale(yscale)

    ax.legend(ncol=1, frameon=False)

    if return_axs:
        return ax
    else:
        return None


def plot_sensitivity(results, mode="violin", per_class=None, max_features = 100, ax=None):
    """Plot results of the sensitivity analysis. They can be plotted in three modes:
    * Violin plot ('violin'), showing the density of per-sample sensitivities besides the mean value
    * Scatter ('scatter'), plotting the mean and standard deviation of the gradients
    * Horizontal bar plot ('barh') only displaying the mean of the gradients

    Parameters
    ----------
    results : dictionary
        sensitity results calculated by sensitivity_analysis
    mode : string, optional
        ('violin','barh','scatter'), by default 'violin'
    per_class : bool, optional
        plot per-class statistics if available, by default plot them if available
    max_features : int, optional
        plot at most max_features, by default 100
    ax : matplotlib axis, optional
        ax where to plot the results, by default it will be initialized

    See also
    --------
    mlcolvar.utils.explain.sensitivity_analysis
        Perform a sensitivity analysis

    Returns
    -------
    ax
        return the generated matplotlib axis if not passed
    """

    # retrieve info from results dictionary
    feature_names = results["feature_names"]
    n_inputs = len(feature_names)
    if max_features < n_inputs:
        print(f'Plotting only the first {max_features} features out of {n_inputs}.')
        feature_names = feature_names[-max_features:]
        n_inputs = max_features
    
    in_num = np.arange(n_inputs)
    n_results = len(results["sensitivity"].keys())

    # check whether to plot per-class statistics
    if per_class is None:
        per_class = True if n_results > 1 else False
    else:
        if not type(per_class) == bool:
            raise TypeError("per_class should be either `True`, `False` or `None`.")
        if per_class & (n_results == 1):
            raise KeyError(
                "Per class analyis requested but no labels found in the results dictionary. You need to call `sensitivity_analysis` with `per_class`=True. "
            )

    # initialize ax
    if ax is None:
        fig = plt.figure(figsize=(5, 0.25 * n_inputs))
        ax = fig.add_subplot(111)
        ax.set_title("Sensitivity Analysis")
        return_ax = True
    else:
        return_ax = False

    # define utils functions
    def _set_violin_attributes(violin_parts, color, alpha=0.5, label=None, zorder=None):
        for pc in violin_parts["bodies"]:
            pc.set_facecolor(color)
            pc.set_edgecolor(color)
            pc.set_alpha(alpha)
            if zorder is not None:
                pc.set_zorder(zorder)
        if label is not None:
            patch_label = (mpatches.Patch(color=color, alpha=alpha), label)
            return patch_label

    patch_labels = []

    patterns = ["", "", "/", "\\", "|", "-", "+", "x", "o", "O", ".", "*"]

    for i, label in enumerate(results["sensitivity"].keys()):
        score = results["sensitivity"][label][-max_features:]
        grad = results["gradients"][label][:,-max_features:]

        color = "fessa0" if "Dataset" in label else f"fessa{7-i}"

        if mode == "barh":
            alpha = 0.6 if "Dataset" in label else 0.4
            height = 0.8 if "Dataset" in label else 0.4
            ax.barh(
                in_num,
                score,
                height=height,
                color=color,
                edgecolor="k",
                hatch=patterns[i],
                alpha=alpha,
                label=label,
            )
        elif mode == "violin":
            widths = 0 if (("Dataset" in label) & per_class) else 0.5
            zorder = 1 if "Dataset" in label else 0
            showmeans = True if "Dataset" in label else False
            violin_parts = ax.violinplot(
                grad,
                positions=in_num,
                vert=False,
                showmeans=showmeans,
                showextrema=False,
                widths=widths,
            )
            patch_label = _set_violin_attributes(
                violin_parts, color, alpha=0.5, label=label, zorder=zorder
            )
            patch_labels.append(patch_label)
            if "Dataset" in label:
                ax.scatter(y=in_num, x=score, c="dimgrey", s=10, zorder=2)
        elif mode == "scatter":
            fmt = "D" if "Dataset" in label else "."
            ax.errorbar(
                y=in_num,
                x=score,
                xerr=grad.std(axis=0),
                color=color,
                fmt=fmt,
                alpha=0.5,
                label=label,
            )
        else:
            raise NotImplementedError(
                'plot mode can be only "barh","violin","scatter".'
            )

        if not per_class:
            break

    # >> legend
    ax.set_xlabel("Sensitivity")
    ax.set_yticks(in_num)
    ax.set_yticklabels(feature_names, fontsize=9)
    ax.set_ylabel("Input features")

    if mode == "violin":
        ax.legend(*zip(*patch_labels), loc="lower right", frameon=False)
    else:
        ax.legend(loc="lower right", frameon=False)
    if np.min(results["sensitivity"]["Dataset"])>=0: 
        ax.set_xlim(0, None)
    else:
        ax.axvline(0,color='grey')
    ax.set_ylim(-1, in_num[-1] + 1)

    if return_ax:
        return ax


def test_utils_plot():
    import matplotlib

    x = np.linspace(-1.5, 1.5)
    y = np.linspace(-0.5, 2.5)

    mp = muller_brown_potential(x, y)
    mp = muller_brown_potential_three_states(x, y)

    pal = paletteFessa
    pal = paletteCortina

    cmap = matplotlib.colors.Colormap("fessa", 2)
    cmap = matplotlib.colors.Colormap("fessa_r", 2)
    cmap = matplotlib.colors.Colormap("cortina80", 2)
    cmap = matplotlib.colors.Colormap("cortina80_r", 2)



#!/usr/bin/env python

# =============================================================================
# MODULE DOCSTRING
# =============================================================================

"""
PyTorch Lightning DataModule object for DictDatasets.
"""

__all__ = ["DictLoader"]


# =============================================================================
# GLOBAL IMPORTS
# =============================================================================

import collections.abc
import math
from typing import Optional, Union, Sequence

import torch
from torch.utils.data import Subset

from mlcolvar.data import DictDataset
from mlcolvar.core.transform.utils import Statistics


# =============================================================================
# FAST DICTIONARY LOADER CLASS
# =============================================================================


class DictLoader:
    """PyTorch DataLoader for :class:`~mlcolvar.data.dataset.DictDataset` .

    It is much faster than ``TensorDataset`` + ``DataLoader`` because ``DataLoader``
    grabs individual indices of the dataset and calls cat (slow).

    The class can also merge multiple :class:`~mlcolvar.data.dataset.DictDataset`s
    that have different keys (see example below). Different datasets can have
    different number of samples. In this case, it is necessary to specify the
    batch sizes so that the number of batches per epoch is the same for all datasets.

    Notes
    -----

    Adapted from https://discuss.pytorch.org/t/dataloader-much-slower-than-manual-batching/27014/6.

    Examples
    --------

    >>> x = torch.arange(1,11)

    A ``DictLoader`` can be initialize from a ``dict``, a :class:`~mlcolvar.data.dataset.DictDataset`,
    or a ``Subset`` wrapping a :class:`~mlcolvar.data.dataset.DictDataset`.

    >>> # Initialize from a dictionary.
    >>> d = {'data': x.unsqueeze(1), 'labels': x**2}
    >>> dataloader = DictLoader(d, batch_size=1, shuffle=False)
    >>> dataloader.dataset_len  # number of samples
    10
    >>> # Print first batch.
    >>> for batch in dataloader:
    ...     print(batch)
    ...     break
    {'data': tensor([[1]]), 'labels': tensor([1])}

    >>> # Initialize from a DictDataset.
    >>> dict_dataset = DictDataset(d)
    >>> dataloader = DictLoader(dict_dataset, batch_size=2, shuffle=False)
    >>> len(dataloader)  # Number of batches
    5

    >>> # Initialize from a PyTorch Subset object.
    >>> train, _ = torch.utils.data.random_split(dict_dataset, [0.5, 0.5])
    >>> dataloader = DictLoader(train, batch_size=1, shuffle=False)

    It is also possible to iterate over multiple dictionary datasets having
    different keys for multi-task learning

    >>> dataloader = DictLoader(
    ...     dataset=[dict_dataset, {'some_unlabeled_data': torch.arange(20)+11}],
    ...     batch_size=[1, 2], shuffle=False,
    ... )
    >>> dataloader.dataset_len  # This is the number of samples in the datasets.
    [10, 20]
    >>>  # Print first batch.
    >>> from pprint import pprint
    >>> for batch in dataloader:
    ...     pprint(batch)
    ...     break
    {'dataset0': {'data': tensor([[1]]), 'labels': tensor([1])},
     'dataset1': {'some_unlabeled_data': tensor([11, 12])}}

    """

    def __init__(
        self,
        dataset: Union[dict, DictDataset, Subset, Sequence],
        batch_size: Union[int, Sequence[int]] = 0,
        shuffle: bool = True,
    ):
        """Initialize a ``DictLoader``.

        Parameters
        ----------
        dataset : dict or DictDataset or Subset of DictDataset or list-like.
            The dataset or a list of datasets. If a list, the datasets can have
            different keys but they must all have the same number of samples.
        batch_size : int or list-like of int, optional
            Batch size, by default 0 (==single batch). If multiple datasets are
            passed, this can be a list specifying the batch size for each dataset.
            Otherwise, if an ``int``, this uses the same batch size for al
            datasets. This must be set so that the total number of batches per
            epoch is the same for all datasets.
        shuffle : bool, optional
            If ``True``, shuffle the data *in-place* whenever an
            iterator is created out of this object, by default ``True``.
        """
        # This checks that dataset and batch_size are consistent.
        self._dataset = None
        self._batch_size = None
        self.set_dataset_and_batch_size(dataset=dataset, batch_size=batch_size)
        self.shuffle = shuffle

        # These are lazily initialized in __iter__().
        self.indices = None
        self.current_batch_idx = None

    @property
    def dataset(self):
        """DictDataset or list[DictDataset]: The dictionary dataset(s)."""
        return self._dataset

    @dataset.setter
    def dataset(self, dataset):
        self.set_dataset_and_batch_size(dataset=dataset, batch_size=None)

    @property
    def has_multiple_datasets(self):
        return not isinstance(self.dataset, DictDataset)

    @property
    def dataset_len(self):
        """int: Number of samples in the dataset(s)."""
        if self.has_multiple_datasets:
            return [len(d) for d in self.dataset]
        return len(self.dataset)

    @property
    def batch_size(self):
        """int or List[int]: Batch size or, in case of multiple datasets, a list of batch sizes."""
        if self.has_multiple_datasets:
            return [
                b if b > 0 else l for b, l in zip(self._batch_size, self.dataset_len)
            ]
        return self._batch_size if self._batch_size > 0 else self.dataset_len

    @batch_size.setter
    def batch_size(self, batch_size):
        self.set_dataset_and_batch_size(dataset=None, batch_size=batch_size)

    @property
    def keys(self):
        """tuple[str] or tuple[tuple[str]]: The keys of all the datasets in this loader."""
        if self.has_multiple_datasets:
            return tuple(d.keys for d in self.dataset)
        return self.dataset.keys

    def set_dataset_and_batch_size(
        self,
        dataset: Union[None, dict, DictDataset, Subset, Sequence],
        batch_size: Union[None, int, Sequence[int]],
    ):
        """Set a compatible pair of datasets and batch sizes.

        With multiple datasets, ``dataset`` and ``batch_size`` must be compatible
        so that each dataset has the same number of batches per epoch so it might
        not be possible to set the two attributes singularly without leaving the
        object in an inconsistent state. Instead, this setter can be used safely.

        Parameters
        ----------
        dataset: None or dict or DictDataset or Subset of DictDataset or list-like.
            The dataset or a list of datasets. If a list, the datasets can have
            different keys but they must all have the same number of samples.
            If ``None``, only ``batch_size`` is set.
        batch_size : None or int or list-like of int
            Batch size, by default 0 (==single batch). If multiple datasets are
            passed, this can be a list specifying the batch size for each dataset.
            Otherwise, if an ``int``, this uses the same batch size for al
            datasets. This must be set so that the total number of batches per
            epoch is the same for all datasets. If ``None``, only ``dataset``
            is set.
        """
        # Save the previous dataset and batch_size. We'll restore them if we find
        # an error to leave the object in a consistent state.
        old_dataset = self._dataset
        old_batch_size = self._batch_size

        if dataset is not None:
            # Convert dicts and Subsets to DictDatasets.
            try:
                dataset = _to_dict_dataset(dataset)
            except ValueError:  # Assume this is a sequence of datasets.
                dataset = [_to_dict_dataset(d) for d in dataset]
            self._dataset = dataset

        # Set batch size.
        if batch_size is not None:
            if self.has_multiple_datasets and not isinstance(
                batch_size, collections.abc.Sequence
            ):
                # If an integer is passed, we set the same batch size to all datasets.
                batch_size = [batch_size] * len(dataset)
            self._batch_size = batch_size

        # Now check for errors.
        if self.has_multiple_datasets:
            # batch_size must have the same length as dataset.
            if len(self._batch_size) != len(self._dataset):
                self._dataset = old_dataset
                self._batch_size = old_batch_size
                raise ValueError(
                    f"batch_size (length {batch_size_len}) must have length equal to the number of datasets (length {len(self.dataset)}."
                )

            # The number of batches per epoch must be the same for all datasets.
            n_batches = [
                math.ceil(dl / b) for dl, b in zip(self.dataset_len, self.batch_size)
            ]
            if len(set(n_batches)) > 1:
                self._dataset = old_dataset
                self._batch_size = old_batch_size
                raise ValueError(
                    "Multiple datasets must have the same number of batches per epoch. "
                    f"With batch_size {self._batch_size} the number of batches are {n_batches}."
                )

    def __iter__(self):
        # Since multiple datasets might have different length, we need to generate
        # separate shuffling indices for all of them.
        if not self.shuffle:
            self.indices = None
        elif self.has_multiple_datasets:
            self.indices = [torch.randperm(l) for l in self.dataset_len]
        else:
            self.indices = torch.randperm(self.dataset_len)

        # Rewind internal batch counter.
        self.current_batch_idx = 0
        return self

    def __next__(self):
        if self.current_batch_idx >= len(self):
            raise StopIteration

        if self.has_multiple_datasets:
            batch = {}
            for dataset_idx in range(len(self.dataset)):
                batch[f"dataset{dataset_idx}"] = self._get_batch(
                    dataset_idx=dataset_idx
                )
        else:
            batch = self._get_batch()

        self.current_batch_idx += 1
        return batch

    def __len__(self):
        """Return the number of batches in the loader."""
        if self.has_multiple_datasets:
            # All datasets have the same number of batches per epoch.
            dataset_len = self.dataset_len[0]
            batch_size = self.batch_size[0]
        else:
            dataset_len = self.dataset_len
            batch_size = self.batch_size
        return (dataset_len + batch_size - 1) // batch_size

    def __repr__(self) -> str:
        string = f"DictLoader(length={self.dataset_len}, batch_size={self.batch_size}, shuffle={self.shuffle})"
        return string

    def get_stats(self, dataset_idx: Optional[int] = None):
        """Compute statistics ``('mean','std','min','max')`` of the dataloader.

        Parameters
        ----------
        dataset_idx : int, optional
            If given and the loader has multiple datasets, only the statistics
            of the ``dataset_idx``-th dataset will be returned.

        Returns
        -------
        stats : Dict[Dict] or List[Dict[Dict]]
            A dictionary mapping the datasets' keys (e.g., ``'data'``, ``'weights'``)
            to their statistics. If the loader has multiple datasets, ``stats[i]``
            is the dictionary for the ``i``-th dataset.

        """
        # Check whether this loader has multiple datasets.
        # Make datasets always a list to simplify the code.
        if self.has_multiple_datasets:
            datasets = self.dataset
        else:
            datasets = [self.dataset]

        # Select requested dataset.
        is_selected_dataset = dataset_idx is not None
        if is_selected_dataset:
            datasets = [datasets[dataset_idx]]

        # Compute stats.
        stats = {f"dataset{i}": {} for i in range(len(datasets))}
        for dataset_idx, dataset in enumerate(datasets):
            for k in dataset.keys:
                stats[f"dataset{dataset_idx}"][k] = Statistics(dataset[k]).to_dict()

        # Return only a single dictionary if there are no multiple datasets.
        if is_selected_dataset or not self.has_multiple_datasets:
            return stats["dataset0"]
        return stats

    def _get_batch(self, dataset_idx=None):
        """Return the current batch from the dataset."""
        # Determine dataset and batch size.
        if dataset_idx is None:  # Only one dataset.
            dataset = self.dataset
            batch_size = self.batch_size
        else:
            dataset = self.dataset[dataset_idx]
            batch_size = self.batch_size[dataset_idx]

        # Determine start and end sample indices.
        start = self.current_batch_idx * batch_size
        end = start + batch_size

        # Handle shuffling.
        if self.indices is None:
            batch = dataset[start:end]
        else:
            if dataset_idx is None:
                indices = self.indices
            else:
                indices = self.indices[dataset_idx]
            batch = dataset[indices[start:end]]

        return batch


# =============================================================================
# PRIVATE UTILITY FUNCTIONS
# =============================================================================


def _to_dict_dataset(d):
    """Convert Dict[Tensor] and Subset[DictDataset] to DictDataset.

    An error is raised if ``d`` cannot is of any other type.
    """
    # Convert to DictDataset if a dict is given.
    if isinstance(d, dict):
        d = DictDataset(d)
    elif isinstance(d, Subset) and isinstance(d.dataset, DictDataset):
        # TODO: This might not not safe for classes that inherit from Subset or DictionaryDatset.
        # Retrieve selection if it a subset.
        d = d.dataset.__class__(d.dataset[d.indices])
    elif not isinstance(d, DictDataset):
        raise ValueError(
            "The data must be of type dict, DictDataset or Subset[DictDataset]."
        )
    return d


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import torch
from mlcolvar.core.transform import Transform


class BaseCV:
    """
    Base collective variable class.

    To inherit from this class, the class must define a BLOCKS class attribute.
    """

    def __init__(
        self,
        in_features,
        out_features,
        preprocessing: torch.nn.Module = None,
        postprocessing: torch.nn.Module = None,
        *args,
        **kwargs,
    ):
        """Base CV class options.

        Parameters
        ----------
        in_features : int
            Number of inputs of the CV model
        out_features : int
            Number of outputs of the CV model, should be the number of CVs
        preprocessing : torch.nn.Module, optional
            Preprocessing module, default None
        postprocessing : torch.nn.Module, optional
            Postprocessing module, default None

        """
        super().__init__(*args, **kwargs)

        # The parent class sets in_features and out_features based on their own
        # init arguments so we don't need to save them here (see #103).
        self.save_hyperparameters(ignore=['in_features', 'out_features'])

        # MODEL
        self.initialize_blocks()
        self.in_features = in_features
        self.out_features = out_features

        # OPTIM
        self._optimizer_name = "Adam"
        self.optimizer_kwargs = {}
        self.lr_scheduler_kwargs = {}

        # PRE/POST
        self.preprocessing = preprocessing
        self.postprocessing = postprocessing

    @property
    def n_cvs(self):
        """Number of CVs."""
        return self.out_features

    @property
    def example_input_array(self):
        return torch.randn(
            self.in_features
            if self.preprocessing is None
            or not hasattr(self.preprocessing, "in_features")
            else self.preprocessing.in_features
        )

    def parse_options(self, options: dict = None):
        """
        Sanitize options and create defaults ({}) if not in options.
        Furthermore, it sets the optimizer kwargs, if given.

        Parameters
        ----------
        options : dict[str, Any], optional
            Options for the building blocks of the model, by default None.
        """
        if options is None:
            options = {}

        for b in self.BLOCKS:
            options.setdefault(b, {})

        for o in options.keys():
            if o not in self.BLOCKS:
                if o == "optimizer":
                    self.optimizer_kwargs.update(options[o])
                elif o == "lr_scheduler":
                    self.lr_scheduler_kwargs.update(options[o])
                else:
                    raise ValueError(
                        f'The key {o} is not available in this class. The available keys are: {", ".join(self.BLOCKS)}, optimizer and lr_scheduler.'
                    )

        return options

    def initialize_blocks(self):
        """
        Initialize the blocks as attributes of the CV class.
        """
        for b in self.BLOCKS:
            self.__setattr__(b, None)

    def setup(self, stage=None):
        if stage == "fit":
            self.initialize_transforms(self.trainer.datamodule)

    def initialize_transforms(self, datamodule):
        for b in self.BLOCKS:
            if isinstance(getattr(self, b), Transform):
                getattr(self, b).setup_from_datamodule(datamodule)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Evaluation of the CV

        - Apply preprocessing if any
        - Execute sequentially all the blocks in self.BLOCKS unless they are not initialized
        - Apply postprocessing if any

        Parameters
        ----------
        x : torch.Tensor
            Input of the forward operation of the model

        Returns
        -------
        torch.Tensor
            Output of the forward operation of the model
        """

        if self.preprocessing is not None:
            x = self.preprocessing(x)

        x = self.forward_cv(x)

        if self.postprocessing is not None:
            x = self.postprocessing(x)

        return x

    def forward_cv(self, x: torch.Tensor) -> torch.Tensor:
        """
        Execute sequentially all the blocks in self.BLOCKS unless they are not initialized.

        No pre/post processing will be executed here. This is supposed to be called during training/validation and to be overloaded if necessary.

        Parameters
        ----------
        x : torch.Tensor
            Input of the forward operation of the model

        Returns
        -------
        torch.Tensor
            Output of the forward operation of the model
        """

        for b in self.BLOCKS:
            block = getattr(self, b)
            if block is not None:
                x = block(x)

        return x

    def validation_step(self, val_batch, batch_idx):
        """
        Equal to training step if not overridden. Different behaviors for train/valid step can be enforced in training_step() based on the self.training variable.
        """
        self.training_step(val_batch, batch_idx)

    def test_step(self, test_batch, batch_idx):
        """
        Equal to training step if not overridden. Different behaviors for train/valid step can be enforced in training_step() based on the self.training variable.
        """
        self.training_step(test_batch, batch_idx)

    @property
    def optimizer_name(self) -> str:
        """Optimizer name. Options can be set using optimizer_kwargs. Actual optimizer will be return during training from configure_optimizer function."""
        return self._optimizer_name

    @optimizer_name.setter
    def optimizer_name(self, optimizer_name: str):
        if not hasattr(torch.optim, optimizer_name):
            raise AttributeError(
                f"torch.optim does not have a {optimizer_name} optimizer."
            )
        self._optimizer_name = optimizer_name

    def configure_optimizers(self):
        """
        Initialize the optimizer based on self._optimizer_name and self.optimizer_kwargs.

        Returns
        -------
        torch.optim
            Torch optimizer
        """

        optimizer = getattr(torch.optim, self._optimizer_name)(
            self.parameters(), **self.optimizer_kwargs
        )

        if self.lr_scheduler_kwargs:
            scheduler_cls = self.lr_scheduler_kwargs['scheduler']
            scheduler_kwargs = {k: v for k, v in self.lr_scheduler_kwargs.items() if k != 'scheduler'}
            lr_scheduler = scheduler_cls(**scheduler_kwargs)
            return [optimizer] , [lr_scheduler]
        else: 
            return optimizer

    def __setattr__(self, key, value):
        # PyTorch overrides __setattr__ to raise a TypeError when you try to assign
        # an attribute that is a Module to avoid substituting the model's component
        # by mistake. This means we can't simply assign to loss_fn a lambda function
        # after it's been assigned a Module, but we need to delete the Module first.
        #    https://github.com/pytorch/pytorch/issues/51896
        #    https://stackoverflow.com/questions/61116433/maybe-i-found-something-strange-on-pytorch-which-result-in-property-setter-not
        try:
            super().__setattr__(key, value)
        except TypeError as e:
            # We make an exception only for loss_fn.
            if (key == "loss_fn") and ("cannot assign" in str(e)):
                del self.loss_fn
                super().__setattr__(key, value)



from lightning import Callback
import copy


class SimpleMetricsCallback(Callback):
    """Lightning callback which append logged metrics to a list.
    The metrics are recorded at the end of each validation epoch.
    """

    def __init__(self):
        super().__init__()
        self.metrics = []

    def on_validation_end(self, trainer, pl_module):
        if not trainer.sanity_checking:
            metrics = copy.deepcopy(trainer.callback_metrics)
            self.metrics.append(metrics)


class MetricsCallback(Callback):
    """Lightning callback which saves logged metrics into a dictionary.
    The metrics are recorded at the end of each validation epoch.
    """

    def __init__(self):
        super().__init__()
        self.metrics = {"epoch": []}

    def on_train_epoch_end(self, trainer, pl_module):
        metrics = trainer.callback_metrics
        if not trainer.sanity_checking:
            self.metrics["epoch"].append(trainer.current_epoch)
            for key, val in metrics.items():
                val = val.item()
                if key in self.metrics:
                    self.metrics[key].append(val)
                else:
                    self.metrics[key] = [val]


def test_metrics_callbacks():
    import torch
    import lightning
    from mlcolvar.cvs import AutoEncoderCV
    from mlcolvar.data import DictDataset, DictModule

    X = torch.rand((100, 2))
    dataset = DictDataset({"data": X})
    datamodule = DictModule(dataset)

    model = AutoEncoderCV([2, 2, 1])
    metrics = SimpleMetricsCallback()
    trainer = lightning.Trainer(
        max_epochs=1,
        log_every_n_steps=2,
        logger=None,
        enable_checkpointing=False,
        callbacks=metrics,
    )
    trainer.fit(model, datamodule)

    model = AutoEncoderCV([2, 2, 1])
    metrics = MetricsCallback()
    trainer = lightning.Trainer(
        max_epochs=1,
        log_every_n_steps=2,
        logger=None,
        enable_checkpointing=False,
        callbacks=metrics,
    )
    trainer.fit(model, datamodule)



# __all__ = ["compute_fes"]

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import torch
import warnings

# optional packages
# pandas
try:
    import pandas as pd

    PANDAS_IS_INSTALLED = True
except ImportError:
    PANDAS_IS_INSTALLED = False
# check whether KDEpy and scikit-learn are installed (used for FES)
try:
    from KDEpy import FFTKDE
    from KDEpy.utils import cartesian

    KDEPY_IS_INSTALLED = True
except ImportError:
    KDEPY_IS_INSTALLED = False
try:
    import sklearn

    SKLEARN_IS_INSTALLED = True
except ImportError:
    SKLEARN_IS_INSTALLED = False
# set default backend based on installation
if KDEPY_IS_INSTALLED:
    kdelib = "KDEpy"
elif SKLEARN_IS_INSTALLED:
    kdelib = "sklearn"
else:
    kdelib = None


def compute_fes(
    X,
    temp=300,
    kbt=None,
    num_samples=100,
    bounds=None,
    bandwidth=0.01,
    kernel="gaussian",
    weights=None,
    scale_by=None,
    blocks=1,
    fes_to_zero=None,
    plot=False,
    plot_max_fes=None,
    ax=None,
    backend=None,
    eps=0,
):
    """Compute the Free Energy Surface along the given variables.

    Parameters
    ----------
    X : array-like, list of arrays
        data
    temp : float, optional
        temperature, by default 300
    kbt : float, optional
        temperature in energy units, by default None
    num_samples : int, optional
        number of points used along each direction, by default 100
    bounds : list of lists, optional
        limit the calculation of the FES to a region, by default equal to the range of values
    bandwidth : float, optional
        Bandwidth use for the kernels, by default 0.1.
    kernel: string, optional
        Kernel type, by default 'gaussian'.
    weights : array, optional
        array of samples weights, by default None
    scale_by : string, optional
        Standardize each variable by its standard deviation (`std`) or by its range of values, by default None
    blocks : int, optional
        number of blocks to be used, by default 1
    fes_to_zero: int, optional
        index of the array where to shift the FES to zero (if none the minimum is set to zero)
    plot : bool, optional
        Plot the results (only available for 1D and 2D FES), by default False
    plot_max_fes : float, optional
        Maximum value of the FES to be plotted, by default None
    ax : matplotlib axis, optional
        Axis where to plot, default create new figure
    backend: string, optional
        Specify the backend for KDE ("KDEpy" or "sklearn")
    eps: float, optional
        Add an epsilon in the argument of the log

    Returns
    -------
    fes: np.array
        free energy
    grid: np.array or list of arrays (>1D)
        grid points for plotting
    bounds: list
        bounds along each variable
    std: np.array
        (weighted) standard deviation along the blocks (only if blocks>1)

    Notes
    -----
    - This function uses sklearn.neighbors.KernelDensity function to construct the estimate of the FES (https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.KernelDensity.html). You can specify all the kernels supported.

    - Note that if ``scale_by`` is not none, the bandwidth is not rescaled as well. So if using ``scale_by='range'`` a bandwidth of 0.01 corresponds to a 1/100 of the range of values assumed by the variable.

    - If blocks>1 a (weighted) block analyis is performed [1], which allows also to obtain an estimate of the uncertainty.

    [1] Invernizzi, Piaggi and Parrinello, PRX, 2020,10,4

    """
    # check backend for KDE
    if kdelib is None:
        raise NotImplementedError(
            "The function compute_fes requires either KDEpy (fast) or scikit-learn (slow) to work."
        )
    if kdelib == "sklearn":
        warnings.warn(
            "Warning: KDEpy is not installed, falling back to scikit-learn for kernel density estimation (very slow for d>1)."
        )

    if backend == "sklearn":
        from sklearn.neighbors import KernelDensity
    elif backend is None:
        backend = kdelib

    # temperature
    if kbt is None:
        kb = 0.00831441
        kbt = kb * temp

    # dataset
    if PANDAS_IS_INSTALLED:
        if type(X) == pd.DataFrame:
            X = X.values
    if type(X) == list:
        X = np.vstack(X).T
    elif type(X) == torch.Tensor:
        X = X.numpy()
    if X.ndim == 1:
        X = X.reshape([-1, 1])

    # div
    nsamples = X.shape[0]
    dim = X.shape[1]

    # weights
    if weights is None:
        weights = np.ones(nsamples)
    else:
        assert weights.ndim == 1
        assert weights.shape[0] == nsamples

    # rescale
    if scale_by is not None:
        if scale_by == "std":
            scale = X.std(axis=0)
        elif scale_by == "range":
            scale = X.max(axis=0) - X.min(axis=0)
        elif type(scale_by) == list:
            scale = np.asarray(scale_by)
        X = np.copy(X)
        X /= scale

    # eval points
    offset = 1e-3
    if dim == 1:
        if bounds is None:
            bounds = (X[:, 0].min() - offset, X[:, 0].max() + offset)
        grid = np.linspace(bounds[0], bounds[1], num_samples)
        positions = grid.reshape([-1, 1])
    else:
        if bounds is None:
            bounds = [
                (X[:, i].min() - offset, X[:, i].max() + offset) for i in range(dim)
            ]
        grid_list = [np.linspace(b[0], b[1], num_samples) for b in bounds]
        grid = np.meshgrid(*grid_list)
        positions = np.vstack([g.ravel() for g in grid]).T

    # divide in blocks
    X_b = np.array_split(X, blocks)
    w_b = np.array_split(weights, blocks)

    O_i, W_i = [], []  # values per block
    # block average
    for i in range(blocks):
        # data of each block
        X_i = X_b[i]
        w_i = w_b[i]

        # fit
        if backend == "KDEpy":
            kde = FFTKDE(bw=bandwidth, kernel=kernel)
            kde.fit(X_i, weights=w_i)

            if dim == 1:
                pos = [grid]
            else:
                pos = grid_list

            # pdf --> fes
            fes = (
                -kbt
                * np.log(kde.evaluate(cartesian(pos)) + eps)
                .reshape([num_samples for i in range(dim)])
                .T
            )

        elif backend == "sklearn":
            kde = KernelDensity(bandwidth=bandwidth, kernel=kernel)
            kde.fit(X_i, sample_weight=w_i)

            # logpdf --> fes
            fes = -kbt * kde.score_samples(positions).reshape(
                [num_samples for i in range(dim)]
            )

        if fes_to_zero is not None:
            fes -= fes[fes_to_zero]
        else:
            fes -= fes.min()

        # result for each block
        O_i.append(fes)
        W_i.append(np.sum(w_i))

    O_i = np.asarray(O_i)
    W_i = np.asarray(W_i)

    # compute avg and std
    if blocks > 1:
        # weighted average
        O = np.dot(O_i.T, W_i) / np.sum(W_i)
        # weighted std
        dev = O_i - O
        blocks_eff = (np.sum(W_i)) ** 2 / (np.sum(W_i**2))
        variance = (
            blocks_eff / (blocks_eff - 1) * (np.dot((dev**2).T, W_i)) / (np.sum(W_i))
        )
        error = np.sqrt(variance / blocks_eff)
    else:
        O = O_i[0]
        error = None

    # rescale back
    if scale_by is not None:
        if dim == 1:
            bounds = (bounds[0] * scale[0], bounds[1] * scale[0])
            grid *= scale
        else:
            bounds = [
                (bounds[i][0] * scale[i], bounds[i][1] * scale[i]) for i in range(dim)
            ]
            for i in range(dim):
                grid[i] *= scale[i]

    # Plot
    if plot:
        if ax is None:
            fig, ax = plt.subplots()
        if dim == 1:
            fes2 = np.copy(O)
            if plot_max_fes is not None:
                fes2[fes2 > plot_max_fes] = np.nan
            if blocks > 1:
                ax.errorbar(grid, fes2, error)
            else:
                ax.plot(grid, fes2)
            ax.set_ylabel("FES")
        elif dim == 2:
            fes2 = np.copy(O)
            if plot_max_fes is not None:
                fes2[fes2 > plot_max_fes] = np.nan
            extent = [item for sublist in bounds for item in sublist]
            pp = ax.contourf(fes2, cmap="fessa", extent=extent)  # ,vmax=max_fes)
            cbar = plt.colorbar(pp, ax=ax)
            cbar.set_label("FES")

    return fes, grid, bounds, error


def test_compute_fes():
    X = np.linspace(1, 11, 100)
    fes, bins, bounds, error_ = compute_fes(
        X=X,
        weights=np.ones_like(X),
        bandwidth=0.02,
        kbt=None,
        num_samples=120,
        bounds=(0, 10),
        fes_to_zero=25,
        scale_by="range",
        blocks=2,
        backend="KDEpy",
    )

    Y = np.random.rand(2, 100)

    if SKLEARN_IS_INSTALLED:  # TODO: change to use pytest functionalities?
        fes, bins, bounds, error_ = compute_fes(
            X=[Y[0], Y[1]],
            weights=np.ones_like(X),
            bandwidth=0.02,
            kbt=1,
            num_samples=120,
            bounds=None,
            fes_to_zero=None,
            scale_by="std",
            blocks=2,
            backend="sklearn",
        )



#!/usr/bin/env python


# =============================================================================
# MODULE DOCSTRING
# =============================================================================

"""
Shared tests for the objects and functions in the mlcolvar.cvs package.
"""


# =============================================================================
# GLOBAL IMPORTS
# =============================================================================

import os
import tempfile

import lightning
import pytest
import torch

import mlcolvar.cvs
from mlcolvar.data import DictDataset, DictModule


# =============================================================================
# GLOBAL VARIABLES
# =============================================================================

N_STATES = 2
N_DESCRIPTORS = 15
LAYERS = [N_DESCRIPTORS, 5, 5, N_STATES-1]

# =============================================================================
# FIXTURES
# =============================================================================

@pytest.fixture(scope="module")
def dataset():
    """Dataset with all fields required by all CV types."""
    n_samples = 10

    # Weights should be optional so we don't add them.
    data = {
        "data": torch.randn((n_samples, N_DESCRIPTORS)),
        "data_lag": torch.randn((n_samples, N_DESCRIPTORS)),
        "target": torch.randn(n_samples),
        "weights": torch.rand(n_samples),
        "weights_lag": torch.rand(n_samples),
    }

    # With sequential sampling, this make sure that all labels are represented
    # in the validation and training set so that LDA/TDA don't complain.
    labels = torch.arange(N_STATES, dtype=torch.get_default_dtype())
    data["labels"] = labels.repeat(n_samples // N_STATES + 1)[:n_samples]

    return DictDataset(data)


# =============================================================================
# TESTS
# =============================================================================

@pytest.mark.parametrize("cv_model", [
    mlcolvar.cvs.DeepLDA(layers=LAYERS, n_states=N_STATES),
    mlcolvar.cvs.DeepTDA(n_states=N_STATES, n_cvs=1, target_centers=[-1., 1.], target_sigmas=[0.1, 0.1], layers=LAYERS),
    mlcolvar.cvs.RegressionCV(layers=LAYERS),
    mlcolvar.cvs.DeepTICA(layers=LAYERS, n_cvs=1),
    mlcolvar.cvs.AutoEncoderCV(encoder_layers=LAYERS),
    mlcolvar.cvs.VariationalAutoEncoderCV(n_cvs=1, encoder_layers=LAYERS[:-1]),
])
def test_resume_from_checkpoint(cv_model, dataset):
    """CVs correctly resume from a checkpoint."""
    datamodule = DictModule(dataset, lengths=[1.0,0.], batch_size=len(dataset))

    # Run a few steps of training in a temporary directory.
    with tempfile.TemporaryDirectory() as tmp_dir_path:
        # Simulate a couple of epochs of training.
        trainer = lightning.Trainer(
            max_epochs=2,
            enable_checkpointing=True,
            logger=False,
            enable_progress_bar=False,
            enable_model_summary=False,
            default_root_dir=tmp_dir_path,
        )
        trainer.fit(cv_model, datamodule)

        # Now load from checkpoint.
        file_name = 'epoch={}-step={}.ckpt'.format(trainer.current_epoch-1, trainer.global_step)
        checkpoint_file_path = os.path.join(tmp_dir_path, 'checkpoints', file_name)
        cv_model2 = cv_model.__class__.load_from_checkpoint(checkpoint_file_path)

    # Check that state is the same.
    x = dataset['data']
    cv_model.eval()
    cv_model2.eval()
    assert torch.allclose(cv_model(x), cv_model2(x))



import numpy as np
import torch

from mlcolvar.utils.plot import plot_sensitivity


def sensitivity_analysis(
    model,
    dataset,
    std=None,
    feature_names=None,
    metric="mean_abs_val",
    per_class=False,
    plot_mode="violin",
    ax=None,
):
    """Perform a sensitivity analysis using the partial derivatives method. This allows us to measure which input features the model is most sensitive to (i.e., which quantities produce significant changes in the output).

    To do this, the partial derivatives of the model with respect to each input :math:`x_i` are computed over a set of `N` points of a :math:`$$\{\mathbf{x}^{(j)}\}_{j=1} ^N$$` dataset.
    These values, in the case where the dataset is not standardized, are multiplied by the standard deviation of the features over the dataset.

    Then, an average sensitivity value :math:`s_i` is computed, either as the mean absolute value (metric=`MAV`):
    .. math:: s_i = \frac{1}{N} \sum_j \left|{\frac{\partial s}{\partial x_i}(\mathbf{x}^{(j)})}\right| \sigma_i

    or as the root mean square (metric=`RMS`):
    .. math:: s_i = \sqrt{\frac{1}{N} \sum_j \left({\frac{\partial s}{\partial x_i}(\mathbf{x}^{(j)})}\  \sigma_i\right)^2 }

    In alternative, one can also compute simply average, without taking the absolute values (metric=`mean`).
    
    In all the above cases, the sensitivity values are normalized such that they sum to 1.

    In case in which a labeled dataset these quantities can be computed also on the subset of the data belonging to each class.

    See also
    --------
    mlcolvar.utils.fes.plot_sensitivity
        Plot the sensitivity analysis results

    Parameters
    ----------
    model : mlcolvar.cvs.BaseCV
        collective variable model
    dataset : mlcovar.data.DictDataset
        dataset on which to compute the sensitivity analysis.
    std : array_like, optional
        standard deviation of the features, by default it will be computed from the dataset
    feature_names : array-like, optional
        array-like with input features names, by default they will be taken from the dataset if available
    metric : str, optional
        sensitivity measure ('mean_abs_val'|'MAV','root_mean_square'|'RMS','mean'), by default 'mean_abs_val'
    per_class : bool, optional
        if the dataset has labels, compute also the sensitivity per class, by default False
    plot_mode : str, optional
        how to visualize the results ('violin','barh','scatter'), by default 'violin'
    ax : matplotlib.axis, optional
        ax where to plot the results, by default it will be initialized

    Returns
    -------
    results: dictionary
        results of the sensitivity analysis, containing 'feature_names', the 'sensitivity' and the 'gradients' per samples, ordered according to the sensitivity.
    """

    # get dataset
    X = dataset["data"]
    n_inputs = X.shape[1]

    # get feature names
    if feature_names is None:
        if dataset.feature_names is not None:
            feature_names = dataset.feature_names
        else:
            feature_names = np.asarray([str(i + 1) for i in range(n_inputs)])

    # get standard deviation
    if std is None:
        std = dataset.get_stats()["data"]["std"].detach().numpy()
    else: 
        std = np.asarray(std)

    # compute cv
    X.requires_grad = True
    s = model(X)

    # get gradients
    grad_output = torch.ones_like(s)
    grad = torch.autograd.grad(s, X, grad_outputs=grad_output)[0].detach().cpu().numpy()
    if metric != "mean":
        grad = np.abs(grad)

    # multiply grad_xi by std_xi
    grad = grad * std

    # normalize such that the average of the abs sums to 1
    grad /= np.abs(grad).mean(axis=0).sum()

    # get metrics
    def _compute_score(grad, metric):
        if (metric == "mean_abs_val") | (metric == "MEAN_ABS") | (metric == "MAV") | (metric == 'mean'):
            score = grad.mean(axis=0)
        elif (metric == "root_mean_square") | (metric == "rms") | (metric == "RMS"):
            score = np.sqrt((grad**2).mean(axis=0))
        else:
            raise NotImplementedError(
                "only `mean_abs_value` (MAV) or `root_mean_square` (RMS), or `mean` metrics are allowed"
            )
        return score

    score = _compute_score(grad, metric)

    # sort features based on (absolute) sensitivity
    index = np.abs(score).argsort()
    feature_names = np.asarray(feature_names)[index]
    score = score[index]
    grad = grad[:, index]

    # store into results
    out = {}
    out["feature_names"] = feature_names
    out["sensitivity"] = {"Dataset": score}
    out["gradients"] = {"Dataset": grad}

    # per class statistics
    if per_class:
        try:
            labels = dataset["labels"].numpy().astype(int)
        except KeyError:
            raise KeyError(
                "Per class analyis requested but no labels found in the given dataset."
            )

        unique_labels = np.unique(labels)
        for i, l in enumerate(unique_labels):
            mask = np.argwhere(labels == l)[:, 0]
            grad_l = grad[mask, :]
            score_l = _compute_score(grad_l, metric)
            out["sensitivity"][f"State {l}"] = score_l
            out["gradients"][f"State {l}"] = grad_l

    # plot
    if plot_mode is not None:
        plot_sensitivity(out, mode=plot_mode, ax=ax)

    return out


def test_sensitivity_analysis():
    from mlcolvar.data import DictDataset
    from mlcolvar.cvs import DeepLDA

    n_states = 2
    in_features, out_features = 2, n_states - 1
    layers = [in_features, 5, 5, out_features]

    # create dataset
    samples = 10
    X = torch.randn((samples * n_states, 2))

    # create labels
    y = torch.zeros(X.shape[0])
    for i in range(1, n_states):
        y[samples * i :] += 1

    dataset = DictDataset({"data": X, "labels": y})

    # define CV
    opts = {
        "nn": {"activation": "shifted_softplus"},
    }
    model = DeepLDA(layers, n_states, options=opts)

    # feature importances
    for per_class in [True, False, None]:
        for names in [None, ["x", "y"], np.asarray(["x", "y"])]:
            results = sensitivity_analysis(
                model, dataset, feature_names=names, per_class=per_class, plot_mode=None
            )



import torch

__all__ = ["Stats"]


class Stats(torch.nn.Module):
    """
    Base stats class.
    To implement a new stats override the compute and forward methods.
    The parameters of the stats should be set either in the initialization or via the setup_from_datamodule function.
    """

    def compute(self, X: torch.Tensor):
        """
        Compute the parameters of the estimator
        """
        raise NotImplementedError

    def forward(self, X: torch.Tensor):
        """
        Apply estimator
        """
        raise NotImplementedError

    def teardown(self):
        pass



"""Time-lagged independent component analysis"""

__all__ = ["TICA"]

import torch
from mlcolvar.core.stats import Stats
from mlcolvar.core.stats.utils import (
    correlation_matrix,
    cholesky_eigh,
    compute_average,
    reduced_rank_eig,
)
from mlcolvar.core.transform.utils import batch_reshape
import warnings


class TICA(Stats):
    """
    Time-lagged independent component analysis base class.
    """

    def __init__(self, in_features, out_features=None):
        """
        Initialize a TICA object.
        """
        super().__init__()

        # save attributes
        self.in_features = in_features
        self.out_features = out_features if out_features is not None else in_features

        # buffers
        # tica eigenvectors
        self.register_buffer("evecs", torch.eye(in_features, self.out_features))
        # mean to obtain mean free inputs
        self.register_buffer("mean", torch.zeros(in_features))

        # init other attributes
        self.evals = None
        self.C_0 = None
        self.C_lag = None

        # Regularization
        self.reg_C_0 = 1e-6

    def extra_repr(self) -> str:
        repr = f"in_features={self.in_features}, out_features={self.out_features}"
        return repr

    def compute(self, data, weights=None, remove_average=True, save_params=False):
        """Perform TICA computation.

        Parameters
        ----------
        data : [list of torch.Tensors]
            Time-lagged configurations (x_t, x_{t+lag})
        weights : [list of torch.Tensors], optional
            Weights at time t and t+lag, by default None
        remove_average: bool, optional
            whether to make the inputs mean free, by default True
        save_params : bool, optional
            Save parameters of estimator, by default False

        Returns
        -------
        torch.Tensor,torch.Tensor
            eigenvalues,eigenvectors

        """
        # parse args

        x_t, x_lag = data
        w_t, w_lag = None, None
        if weights is not None:
            w_t, w_lag = weights

        if remove_average:
            x_ave = compute_average(x_t, w_t)
            x_t = x_t.sub(x_ave)
            x_lag = x_lag.sub(x_ave)

        C_0 = correlation_matrix(x_t, x_t, w_t)
        C_lag = correlation_matrix(x_t, x_lag, w_lag)

        evals, evecs = cholesky_eigh(C_lag, C_0, self.reg_C_0, n_eig=self.out_features)

        if save_params:
            self.evals = evals
            self.evecs = evecs
            if remove_average:
                self.mean = x_ave

        return evals, evecs

    def timescales(self, lag):
        r"""Return implied timescales from eigenvalues and lag-time.

        Parameters
        ----------
        lag : float
            lag-time

        Returns
        -------
        its : tensor
            implied timescales

        Notes
        -----
        If `lambda_i` are the eigenvalues and `tau` the lag-time, the implied times are given by:

        .. math:: t_i = - tau / \log\lambda_i
        """

        its = -lag / torch.log(torch.abs(self.evals))

        return its

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Compute linear combination with saved eigenvectors

        Parameters
        ----------
        x: torch.Tensor
            input

        Returns
        -------
        out : torch.Tensor
            output
        """
        mean = batch_reshape(self.mean, x.size())

        return torch.matmul(x.sub(mean), self.evecs)


def test_tica():
    in_features = 2
    X = torch.rand(100, in_features) * 100
    x_t = X[:-1]
    x_lag = X[1:]
    w_t = torch.rand(len(x_t))
    w_lag = w_t

    # direct way, compute tica function
    tica = TICA(in_features, out_features=2)
    print(tica)
    tica.compute([x_t, x_lag], [w_t, w_lag], save_params=True)
    s = tica(X)
    print(X.shape, "-->", s.shape)
    print("eigvals", tica.evals)
    print("timescales", tica.timescales(lag=10))

    # step by step
    tica = TICA(in_features)
    C_0 = correlation_matrix(x_t, x_t)
    C_lag = correlation_matrix(x_t, x_lag)
    print(C_0.shape, C_lag.shape)

    evals, evecs = cholesky_eigh(C_lag, C_0)
    print(evals.shape, evecs.shape)

    print(">> batch")
    s = tica(X)
    print(X.shape, "-->", s.shape)
    print(">> single")
    X2 = X[0]
    s2 = tica(X2)
    print(X2.shape, "-->", s2.shape)


if __name__ == "__main__":
    test_tica()



#!/usr/bin/env python

# =============================================================================
# MODULE DOCSTRING
# =============================================================================

"""
Evidence Lower BOund (ELBO) loss functions used to train variational Autoencoders.
"""

__all__ = ["ELBOGaussiansLoss", "elbo_gaussians_loss"]


# =============================================================================
# GLOBAL IMPORTS
# =============================================================================

from typing import Optional
import torch
from mlcolvar.core.loss.mse import mse_loss


# =============================================================================
# LOSS FUNCTIONS
# =============================================================================


class ELBOGaussiansLoss(torch.nn.Module):
    """ELBO loss function assuming the latent and reconstruction distributions are Gaussian.

    The ELBO uses the MSE as the reconstruction loss (i.e., assumes that the
    decoder outputs the mean of a Gaussian distribution with variance 1), and
    the KL divergence between two normal distributions ``N(mean, var)`` and
    ``N(0, 1)``, where ``mean`` and ``var`` are the output of the encoder.
    """

    def forward(
        self,
        target: torch.Tensor,
        output: torch.Tensor,
        mean: torch.Tensor,
        log_variance: torch.Tensor,
        weights: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """Compute the value of the loss function.

        Parameters
        ----------
        target : torch.Tensor
            Shape ``(n_batches, in_features)``. Data points (e.g. input of encoder
            or time-lagged features).
        output : torch.Tensor
            Shape ``(n_batches, in_features)``. Output of the decoder.
        mean : torch.Tensor
            Shape ``(n_batches, latent_features)``. The means of the Gaussian
            distributions associated to the inputs.
        log_variance : torch.Tensor
            Shape ``(n_batches, latent_features)``. The logarithm of the variances
            of the Gaussian distributions associated to the inputs.
        weights : torch.Tensor, optional
            Shape ``(n_batches,)`` or ``(n_batches,1)``. If given, the average over
            batches is weighted. The default (``None``) is unweighted.

        Returns
        -------
        loss: torch.Tensor
            The value of the loss function.
        """
        return elbo_gaussians_loss(target, output, mean, log_variance, weights)


def elbo_gaussians_loss(
    target: torch.Tensor,
    output: torch.Tensor,
    mean: torch.Tensor,
    log_variance: torch.Tensor,
    weights: Optional[torch.Tensor] = None,
) -> torch.Tensor:
    """ELBO loss function assuming the latent and reconstruction distributions are Gaussian.

    The ELBO uses the MSE as the reconstruction loss (i.e., assumes that the
    decoder outputs the mean of a Gaussian distribution with variance 1), and
    the KL divergence between two normal distributions ``N(mean, var)`` and
    ``N(0, 1)``, where ``mean`` and ``var`` are the output of the encoder.

    Parameters
    ----------
    target : torch.Tensor
        Shape ``(n_batches, in_features)``. Data points (e.g. input of encoder
        or time-lagged features).
    output : torch.Tensor
        Shape ``(n_batches, in_features)``. Output of the decoder.
    mean : torch.Tensor
        Shape ``(n_batches, latent_features)``. The means of the Gaussian
        distributions associated to the inputs.
    log_variance : torch.Tensor
        Shape ``(n_batches, latent_features)``. The logarithm of the variances
        of the Gaussian distributions associated to the inputs.
    weights : torch.Tensor, optional
        Shape ``(n_batches,)`` or ``(n_batches,1)``. If given, the average over
        batches is weighted. The default (``None``) is unweighted.

    Returns
    -------
    loss: torch.Tensor
        The value of the loss function.
    """
    # KL divergence between N(mean, variance) and N(0, 1).
    # See https://stats.stackexchange.com/questions/7440/kl-divergence-between-two-univariate-gaussians
    kl = -0.5 * (log_variance - log_variance.exp() - mean**2 + 1).sum(dim=1)

    # Weighted mean over batches.
    if weights is None:
        kl = kl.mean()
    else:
        weights = weights.squeeze()
        if weights.shape != kl.shape:
            raise ValueError(
                f"weights should be a tensor of shape (n_batches,) or (n_batches,1), not {weights.shape}."
            )
        kl = (kl * weights).sum()

    # Reconstruction loss.
    reconstruction = mse_loss(output, target, weights=weights)

    return reconstruction + kl



import torch
import lightning
from mlcolvar.cvs import BaseCV
from mlcolvar.core import FeedForward, Normalization
from mlcolvar.core.loss import MSELoss

__all__ = ["RegressionCV"]


class RegressionCV(BaseCV, lightning.LightningModule):
    """
    Example of collective variable obtained with a regression task.
    Combine the inputs with a neural-network and optimize it to match a target function.

    **Data**: for training it requires a DictDataset with the keys 'data' and 'target' and optionally 'weights'.

    **Loss**: least squares (MSELoss).

    See also
    --------
    mlcolvar.core.loss.MSELoss
        (weighted) Mean Squared Error (MSE) loss function.
    """

    BLOCKS = ["norm_in", "nn"]

    def __init__(self, layers: list, options: dict = None, **kwargs):
        """Example of collective variable obtained with a regression task.
        By default a module standardizing the inputs is used.

        Parameters
        ----------
        layers : list
            Number of neurons per layer
        options : dict[str, Any], optional
            Options for the building blocks of the model, by default None.
            Available blocks: ['norm_in', 'nn'].
            Set 'block_name' = None or False to turn off that block
        """
        super().__init__(in_features=layers[0], out_features=layers[-1], **kwargs)

        # =======   LOSS  =======
        self.loss_fn = MSELoss()

        # ======= OPTIONS =======
        # parse and sanitize
        options = self.parse_options(options)

        # Initialize norm_in
        o = "norm_in"
        if (options[o] is not False) and (options[o] is not None):
            self.norm_in = Normalization(self.in_features, **options[o])

        # initialize NN
        o = "nn"
        self.nn = FeedForward(layers, **options[o])

    def training_step(self, train_batch, batch_idx):
        """Compute and return the training loss and record metrics."""
        # =================get data===================
        x = train_batch["data"]
        labels = train_batch["target"]
        loss_kwargs = {}
        if "weights" in train_batch:
            loss_kwargs["weights"] = train_batch["weights"]
        # =================forward====================
        y = self.forward_cv(x)
        # ===================loss=====================
        loss = self.loss_fn(y, labels, **loss_kwargs)
        # ====================log=====================
        name = "train" if self.training else "valid"
        self.log(f"{name}_loss", loss, on_epoch=True)
        return loss


def test_regression_cv():
    """
    Create a synthetic dataset and test functionality of the RegressionCV class
    """
    from mlcolvar.data import DictDataset, DictModule

    in_features, out_features = 2, 1
    layers = [in_features, 5, 10, out_features]

    # initialize via dictionary
    options = {"nn": {"activation": "relu"}}

    model = RegressionCV(layers=layers, options=options)
    print("----------")
    print(model)

    # create dataset
    X = torch.randn((100, 2))
    y = X.square().sum(1)
    dataset = DictDataset({"data": X, "target": y})
    datamodule = DictModule(dataset, lengths=[0.75, 0.2, 0.05], batch_size=25)
    # train model
    model.optimizer_name = "SGD"
    model.optimizer_kwargs.update(dict(lr=1e-2))
    trainer = lightning.Trainer(
        accelerator="cpu", max_epochs=1, logger=None, enable_checkpointing=False
    )
    trainer.fit(model, datamodule)
    model.eval()
    # trace model
    traced_model = model.to_torchscript(
        file_path=None, method="trace", example_inputs=X[0]
    )
    assert torch.allclose(model(X), traced_model(X))

    # weighted loss
    print("weighted loss")
    w = torch.randn((100))
    dataset_weights = DictDataset({"data": X, "target": y, "weights": w})
    datamodule_weights = DictModule(
        dataset_weights, lengths=[0.75, 0.2, 0.05], batch_size=25
    )
    trainer.fit(model, datamodule_weights)

    # use custom loss
    print("custom loss")
    trainer = lightning.Trainer(
        accelerator="cpu", max_epochs=1, logger=None, enable_checkpointing=False
    )

    model = RegressionCV(layers=[2, 10, 10, 1])
    model.loss_fn = lambda y, y_ref: (y - y_ref).abs().mean()
    trainer.fit(model, datamodule)


if __name__ == "__main__":
    test_regression_cv()
